import os
import math
import pathlib
from . import tree
from .attributes import Attributes
from .util import polytool
from .util import cache
from .util import hash
from .util import vmath




class Solid:
    """
    Informational base class of :ref:`Solid<Solid>` objects. This empty class is implemented for type matching purposes.
    """
    def __init__ (self):
        #: Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = 3




class Sphere(tree.Item, Solid):
    """
    A Sphere is a basic :ref:`Solid<Solid>` class.
    
    The Sphere can be definied with the parameter *radius* or *diameter*. Only the radius will be stored.
    The Spheres center is at coordinate (0, 0, 0).
    """
    def __init__ (self,
                    radius: float = None,
                    diameter: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid arguments
        assert (radius != None) or (diameter != None), "Expected parameter radius or diameter"
        assert (radius == None) or (diameter == None), "Parameters radius and diameter can not be used together"
        if diameter != None:
            radius = diameter / 2

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = 3

        #: Radius of the sphere.
        self.radius = radius

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'radius'
        )




class Cube(tree.Item, Solid):
    """
    A Cube is a basic :ref:`Solid<Solid>` class.
    
    The Cube is defined by a three dimensional size tuple.
    The Cubes center is at coordinate (0, 0, 0).
    """
    def __init__ (self,
                    size: tuple = (1, 1, 1),
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid size parameter
        assert size != None, "Expected parameter size to be defined"
        if isinstance (size, (int, float)):
            size = (size, size, size)
        elif isinstance (size, list):
            size = tuple (size)
        elif not isinstance (size, tuple):
            assert False, "Expected size to be a number or a 3 dimensional vector"
        assert len (size) == 3, "Expected size to be a number or a 3 dimensional vector"
        assert isinstance (size[0], (int, float)), "Expected size to be a tuple containing 3 numbers"
        assert isinstance (size[1], (int, float)), "Expected size to be a tuple containing 3 numbers"
        assert isinstance (size[2], (int, float)), "Expected size to be a tuple containing 3 numbers"

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = 3

        #: Size of Cube as tuple containing (width: float, height: float, depth: float).
        self.size = size

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'size'
        )    




class Cylinder(tree.Item, Solid):
    """
    A Cylinder is a basic :ref:`Solid<Solid>` class.
    
    The Cylinder is defined by a height and two diameters or radius values.
    The Cylinders center is at coordinate (0, 0, 0).
    """
    def __init__ (self,
                    height: float = 1,
                    radius1: float = None,
                    radius2: float = None,
                    diameter1: float = None,
                    diameter2: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check parameters for height
        assert isinstance (height, (int, float)), "Expected a number as parameter height"

        # check parameters for radius1
        if radius1 != None:
            assert isinstance (radius1, (int, float)), "Expected a number as parameter radius1"
            assert diameter1 == None, "Parameter radius1 can not be used together with diameter1"
        elif diameter1 != None:
            assert isinstance (diameter1, (int, float)), "Expected a number as parameter diameter1"
            radius1 = diameter1 / 2
        else:
            radius1 = 1

        # check parameters for radius2
        if radius2 != None:
            assert isinstance (radius2, (int, float)), "Expected a number as parameter radius2"
            assert diameter2 == None, "Parameter radius2 can not be used together with diameter2"
        elif diameter2 != None:
            assert isinstance (diameter2, (int, float)), "Expected a number as parameter diameter2"
            radius2 = diameter2 / 2
        else:
            radius2 = radius1

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = 3

        #: Height of the Cylinder.
        self.height = height

        #: Radius of Cylinder at z = **-height** / 2.
        self.radius1 = radius1

        #: Radius of Cylinder at z = **height** / 2.
        self.radius2 = radius2

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'height',
            'radius1',
            'radius2'
        )    




class Polyhedron(tree.Item, Solid):
    """
    A Polyhedron is a :ref:`Solid<Solid>` formed by a closed surface of triangles.
    """
    def __init__ (self,
                    points: tuple = None,
                    faces: tuple = None,
                    normals: tuple = None,
                    uvvectors: tuple = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        super ().__init__ (name = name, attributes = attributes)

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = 3

        #: Tuple containing the three dimensional position vectors used for declaraing the corners of a face.
        self.points = points

        #: Tuple containing the three dimensional normal vectors used for defining the normals of a face.
        #: Normals are optional.
        self.normals = normals

        #: Tuple containing the two dimensional uv vectors used for defining the texture mapping of a face.
        #: Uv vectors are optional.
        self.uvvectors = uvvectors

        #: A Polyhedron is definied by a closed set of faces. Each face describes a triangle with optional
        #: normals and texture mappings for each point. Faces are references to points, normals and uv vectors
        #: in the format (p1, p2, p3[, n1, n2, n3 [, vu1, vu2, vu3]]). A face must have 3, 6 or 9 elements.
        #: Normal and vu vector indices are optional and can be set to None if a point doesn't have a normal
        #: or uv mapping.
        self.faces = faces


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'points',
            'faces',
            'normals',
            'uvvectors'
        )


    @staticmethod
    def _removeDuplicatesWithIndices (points, faces, normals, uvs):
        """
        Optimize polyhedron: remove duplicated points, normals and uv vectors, adjusts face indices accodingly.
        """
        # make point, normals, uvs unique
        newPoints, pointMap = polytool._removeDuplicatesVector3 (points)
        newNormals, normalMap = polytool._removeDuplicatesVector3 (normals)
        newUvs, uvsMap = polytool._removeDuplicatesVector2 (uvs)

        # replace indices in faces
        newFaces = []
        for face in faces:
            newFace = []
            if len (face) >= 3:
                # process points
                newFace.append (pointMap [face[0]])
                newFace.append (pointMap [face[1]])
                newFace.append (pointMap [face[2]])
            if len (face) >= 6:
                # process normals
                newFace.append (normalMap [face[3]] if face[3] != None else None)
                newFace.append (normalMap [face[4]] if face[4] != None else None)
                newFace.append (normalMap [face[5]] if face[5] != None else None)
            if len (face) >= 9:
                # process uvs
                newFace.append (uvsMap [face[6]] if face[6] != None else None)
                newFace.append (uvsMap [face[7]] if face[7] != None else None)
                newFace.append (uvsMap [face[8]] if face[8] != None else None)
            elif len (face) > 9:
                assert False, "unsupported face format"
            newFaces.append (newFace)

        # return result
        return (newPoints, newFaces, newNormals, newUvs)


    @staticmethod
    def fromSolid (solid: tree.Item, rasterizingAttributes: Attributes = None, name: str = None, attributes: Attributes = None):
        """
        Create a Polyhedron by rasterizing a three dimensional csg tree :ref:`Item<Item>`.
        The accuracy for rasterizing is calculated from *rasterizingAttributes*.
        When the solid has disjunct volumes, a :ref:`Group<Group>` of Polyhedrons will be returned.
        When the solid has no renderable items, :ref:`Empty<Empty>` will be returned.
        """
        rasterized = _toPolyhedronsWrapper (solid, rasterizingAttributes)
        return rasterized.copy (name = name, attributes = attributes)


    @staticmethod
    def _parseFloatArgs (line, count = None):
        """
        Parse array of numbers.
        """
        l = line.split (" ")
        r = []
        for e in l:
            if e != "":
                r.append (float (e))
        return r


    @staticmethod
    def _parseIntegerArgs (line, count = None):
        """
        Parse array of integers.
        """
        l = line.split (" ")
        r = []
        for e in l:
            if e != "":
                r.append (int (e))
        return r


    @staticmethod
    def _checkIsCCW (normalVector, facetPoints, default):
        center = vmath.vsDivScalar (vmath.vsAdd (vmath.vsAdd (facetPoints[0], facetPoints[1]), facetPoints[2]), 3)
        s = vmath.vsAdd (center, normalVector)
        va = vmath.vsSub (facetPoints[1], facetPoints[0])
        vb = vmath.vsSub (facetPoints[2], facetPoints[0])
        vc = vmath.vsSub (s, facetPoints[0])
        axb = (
            va[1] * vb[2] - va[2] * vb[1],
            va[2] * vb[0] - va[0] * vb[2],
            va[0] * vb[1] - va[1] * vb[0]
        )
        vd = axb[0] * vc[0] + axb[1] * vc[1] + axb[2] * vc[2]
        if vd > 0:
            return True
        elif vd < 0:
            return False
        else:
            return default


    def _fromSubsetOfFaces (self, facesIncluded, name, attributes):
        """
        Returns a new Polyhedron consisting of a subset of the current faces
        """
        newPoints = []
        pointRemap = {}
        newNormals = []
        normalRemap = {}
        newVus = []
        vuRemap = {}
        newFaces = []

        # process each face
        for faceId in facesIncluded:
            face = self.faces[faceId]
            newFace = []
            # process points...
            for index in range (0, 3):
                pid = face[index]
                if pid in pointRemap:
                    npid = pointRemap[pid]
                else:
                    npid = len (newPoints)
                    newPoints.append (self.points[pid])
                    pointRemap[pid] = npid
                newFace.append (npid)
            if len (face) > 3:
                # process normal vectors
                for index in range (3, 6):
                    pid = face[index] if index < len (face) else None
                    if pid in normalRemap:
                        npid = normalRemap[pid]
                    else:
                        npid = len (newNormals)
                        newNormals.append (self.normals[pid])
                        normalRemap[pid] = npid
                    newFace.append (npid)
            if len (face) > 6:
                # process vu vectors
                for index in range (6, 9):
                    pid = face[index] if index < len (face) else None
                    if pid in vuRemap:
                        npid = vuRemap[pid]
                    else:
                        npid = len (newVus)
                        newVus.append (self.uvvectors[pid])
                        vuRemap[pid] = npid
                    newFace.append (npid)
            # append new face
            newFaces.append (newFace)

        # return subset
        return Polyhedron (newPoints, newFaces, newNormals, newVus, name, attributes)


    @staticmethod
    def _splitCollectFaces (pointsToFaces, facesToMashMap, faceIndex, currentSolidIdentifier, volumeFaces, faceList):
        # start on face index
        delayedProcess = [faceIndex]
        alreadyProcessed = set ()
        alreadyProcessed.add (faceIndex)
        
        # delay processed nodes
        while len (delayedProcess) > 0:
            # pop from delayed process stack
            popDelay = delayedProcess.pop ()

            # add node to face list and flag as used
            faceList.append (popDelay)
            facesToMashMap[popDelay] = currentSolidIdentifier

            # process neighbours
            face = volumeFaces[popDelay]
            for edge in range (0, 3):
                pid1 = face[edge % 3]
                pid2 = face[(edge + 1) % 3]
                pid = str (pid1) + "," + str (pid2) if pid1 < pid2 else str (pid2) + "," + str (pid1)
                for anchestor in pointsToFaces[pid]:
                    if not anchestor in alreadyProcessed:
                        alreadyProcessed.add (anchestor)
                        delayedProcess.append (anchestor)


    @staticmethod
    def _splitVolume (volume):
        # create point to face map and solid mapping dictionary
        pointsToFaces = {}
        facesToMashMap = []
        for faceId in range (0, len (volume.faces)):
            face = volume.faces[faceId]
            for edge in range (0, 3):
                pid1 = face[edge % 3]
                pid2 = face[(edge + 1) % 3]
                pid = str (pid1) + "," + str (pid2) if pid1 < pid2 else str (pid2) + "," + str (pid1)
                if pid in pointsToFaces:
                    pointsToFaces[pid].append (faceId)
                else:
                    pointsToFaces[pid] = [faceId]
            facesToMashMap.append (None)

        # split algorithm
        splitVolumes = []
        currentSolidIdentifier = -1
        nextFaceToCheck = 0
        while nextFaceToCheck < len (facesToMashMap):
            # when already marked, skip node
            if facesToMashMap[nextFaceToCheck] != None:
                nextFaceToCheck += 1

            # face is not marked yet, pick a new solid identifier and mark it.
            else:
                currentSolidIdentifier += 1
                faceList = []
                Polyhedron._splitCollectFaces (pointsToFaces, facesToMashMap, nextFaceToCheck, currentSolidIdentifier, volume.faces, faceList)
                splitVolumes.append (faceList)
                facesToMashMap[nextFaceToCheck] = currentSolidIdentifier
                nextFaceToCheck += 1

        # generate result by splitting volumes
        if len (splitVolumes) > 1:
            items = []
            solidId = 0
            for faces in splitVolumes:
                newName = volume.name
                if newName in (None, ""):
                    newName = str (solidId)
                else:
                    newName = newName + "-" + str (solidId)
                solidId += 1
                items.append (volume._fromSubsetOfFaces (faces, volume.name, volume.attributes))
            return items

        # return solid as it is
        return [volume]


    @staticmethod
    def _splitVolumes (inputs, name: str, attributes: Attributes):
        """
        Split parsed volumes.
        """
        # collect and split volumes
        volumes = []
        if inputs == None:
            return None
        if isinstance (inputs, tree.Group):
            for child in inputs.children:
                volumes = volumes + Polyhedron._splitVolume (child)
        else:
            volumes = volumes + Polyhedron._splitVolume (inputs)

        # write back result
        if len (volumes) == 0:
            return None
        if len (volumes) == 1:
            return volumes[0]
        return tree.Group (children = volumes, name = name, attributes = attributes)


    @staticmethod
    def _loadStlFile (fileName, name: str = None, attributes: Attributes = None):
        """
        Load stl file.
        """
        correctOrientation = True

        # storage for parsed points and faces
        parsedPoints = []
        faces = []
        parsedSolids = []

        # read file
        with open (fileName) as f:
            state = 0
            normalVector = None
            facetPoints = []
            for line in f:
                # remove comments and strip line
                line = line.split (';', 1)[0].strip ().lower ()
                if line != "":
                    if state == 3:
                        # parse vertex coordinates
                        if line == "endloop":
                            assert len (facetPoints) == 3, "Expected 3 points per facet"
                            if correctOrientation:
                                isCCW = Polyhedron._checkIsCCW (normalVector, facetPoints, True)
                            else:
                                isCCW = True
                            firstPointIndex = len (parsedPoints)
                            if isCCW:
                                parsedPoints.append ((facetPoints[0], firstPointIndex))
                                parsedPoints.append ((facetPoints[1], firstPointIndex + 1))
                                parsedPoints.append ((facetPoints[2], firstPointIndex + 2))
                            else:
                                parsedPoints.append ((facetPoints[1], firstPointIndex))
                                parsedPoints.append ((facetPoints[0], firstPointIndex + 1))
                                parsedPoints.append ((facetPoints[2], firstPointIndex + 2))
                            faces.append ((firstPointIndex, firstPointIndex + 1, firstPointIndex + 2))
                            state = 4
                        elif line.startswith ("vertex "):
                            rest = line[7:].strip ()
                            coordinates = Polyhedron._parseFloatArgs (rest)
                            assert len (coordinates) == 3, "Expected 3 coodinates per normal"
                            facetPoints.append (coordinates)
                        else:
                            assert False, "expected endloop or vertex keyword"
    
                    elif state == 4:
                        # expect endfacet statement
                        if line == "endfacet":
                            state = 1
                        else:
                            assert False, "expected endfacet keyword"

                    elif state == 2:
                        # expect outer loop statement
                        if line.startswith ("outer "):
                            rest = line[6:].strip ()
                            if rest == "loop":
                                state = 3
                            else:
                                assert False, "expected loop keyword"
                        else:
                            assert False, "expected outer keyword"

                    elif state == 1:
                        # expect a facet or end of solid
                        normalVector = None
                        facetPoints = []
                        if line.startswith ("endsolid ") or line == "endsolid":
                            # finish currently parsed solid
                            optPoints, optFaces, optNormals, optVUs = Polyhedron._removeDuplicatesWithIndices (parsedPoints, faces, None, None)
                            poly = Polyhedron (optPoints, optFaces, optNormals, optVUs, name, attributes)
                            parsedSolids.append (poly)
                            parsedPoints = []
                            faces = []
                            state = 5
                        elif line == "facet":
                            state = 2
                        elif line.startswith ("facet "):
                            rest = line[6:].strip ()
                            if rest != "":
                                if rest.startswith ("normal "):
                                    rest = rest[7:].strip ()
                                    normalVector = Polyhedron._parseFloatArgs (rest)
                                    assert len (normalVector) == 3, "Expected 3 coodinates per normal"
                                else:
                                    assert False, "expected normal keyword"
                            state = 2
                        else:
                            assert False, "expected facet declaration or endsolid keyword"

                    elif state == 0:
                        # expect a solid keyword
                        if line.startswith ("solid ") or line == "solid":
                            state = 1
                        else:
                            assert False, "expected solid keyword"

                    elif state == 5:
                        if line.startswith ("solid ") or line == "solid":
                            # allow to start on next solid.
                            state = 1
                        else:
                            # expected end of file
                            assert False, "expected end of file"
            assert state == 5, "unexpected end of file"

        # empty list of parsed solids?
        if len (parsedSolids) == 0:
            result = tree.Empty (name, attributes)
        elif len (parsedSolids) == 1:
            result = parsedSolids[0]
        else:
            result = tree.Group (children = parsedSolids, name = name, attributes = attributes)

        # split volumes
        return Polyhedron._splitVolumes (result, name, attributes)


    @staticmethod
    def _loadOffFile (fileName, name: str = None, attributes: Attributes = None):
        """
        Load off file.
        """
        # read file
        with open (fileName) as f:
            state = 0
            points = []
            faces = []
            pointsLeft = None
            facesLeft = None
            for line in f:
                # remove comments and strip line
                line = line.split ('#', 1)[0].strip ()
                if line != "":
                    if state == 2:
                        # parse points
                        values = Polyhedron._parseFloatArgs (line.strip ())
                        assert len (values) == 3, "expected 3 coordinates on point definition"
                        points.append (((values[0], values[1], values[2]), len (points)))
                        pointsLeft -= 1
                        if pointsLeft == 0:
                            state = 3

                    elif state == 3:
                        # parse faces
                        values = Polyhedron._parseIntegerArgs (line.strip ())
                        assert len (values) >= 4, "expected at least 4 values on face definition"
                        vcount = values[0]
                        assert vcount >= 3, "expected at least 3 points on a face definition"
                        fp1 = values[1]
                        fp2 = values[2]
                        fp = values[3]
                        assert fp1 < len (points), "invalid point index"
                        assert fp2 < len (points), "invalid point index"
                        assert fp < len (points), "invalid point index"
                        faces.append ((fp1, fp2, fp))
                        for fpid in range (4, len (values)):
                            fc = values[fpid]
                            assert fc < len (points), "invalid point index"
                            faces.append ((fp1, fp, fc))
                            fp = fc

                        if False:
                            assert values[0] in (3, 4), "expected 3 or 4 points on face definition"
                            assert values[1] < len (points), "invalid point index"
                            assert values[2] < len (points), "invalid point index"
                            assert values[3] < len (points), "invalid point index"
                            faces.append ((values[1], values[2], values[3]))

                        facesLeft -= 1
                        if facesLeft == 0:
                            state = 4

                    elif state == 0:
                        # try to parse file type identifier
                        if line == "OFF":
                            state = 1
                        else:
                            # size definition on first line?
                            values = Polyhedron._parseIntegerArgs (line[3:].strip ())
                            assert len (values) in (2, 3), "invalid size definition"
                            pointsLeft = values[0]
                            facesLeft = values[1]
                        assert pointsLeft > 0, "number of points must be greater zero"
                        assert facesLeft > 0, "number of faces must be greater zero"
                        state = 2

                    elif state == 1:
                        # try to parse number of points and vertexes
                        values = line[3:].strip ().split (" ")
                        assert (len (values) >= 2), "invalid size definition"
                        pointsLeft = int (values[0])
                        facesLeft = int (values[1])
                        assert pointsLeft > 0, "number of points must be greater zero"
                        assert facesLeft > 0, "number of faces must be greater zero"
                        state = 2

                    elif state == 4:
                        assert False, "extra content not expected"

        # abort if no data was parsed
        if len (faces) == 0:
            return tree.Empty (name, attributes)

        # remove duplicates from points
        optPoints, optFaces, optNormals, optUVs = Polyhedron._removeDuplicatesWithIndices (points, faces, None, None)
        poly = Polyhedron (optPoints, optFaces, None, None, name, attributes)

        # split volumes
        return Polyhedron._splitVolumes (poly, name, attributes)



    @staticmethod
    def fromFileOFF (fileName, name: str = None, attributes: Attributes = None):
        """
        Create a Polyhedron by reading an OFF file.
        When the imported solid has disjunct volumes, a :ref:`Group<Group>` of Polyhedrons will be returned.
        When the imported solid has no renderable items, :ref:`Empty<Empty>` will be returned.
        """
        # hash code of file
        fileCreationTime = os.path.getctime (fileName)
        fileSize = os.path.getsize (fileName)
        fileHash = hash (fileName, "Polyhedron from file", fileCreationTime, fileSize)

        # try to load from cache and return with attributes
        loaded = cache.load (fileHash, ".phed")
        if loaded != None:
            return loaded.copy (name = name, attributes = attributes)

        # load off file
        generated = Polyhedron._loadOffFile (fileName, name, attributes)
        cache.store (generated, fileHash, ".phed")
        return generated


    @staticmethod
    def fromFileSTL (fileName, name: str = None, attributes: Attributes = None):
        """
        Create a Polyhedron by reading a STL file.
        When the imported solid has disjunct volumes, a :ref:`Group<Group>` of Polyhedrons will be returned.
        When the imported solid has no renderable items, :ref:`Empty<Empty>` will be returned.
        """
        # hash code of file
        fileCreationTime = os.path.getctime (fileName)
        fileSize = os.path.getsize (fileName)
        fileHash = hash (fileName, "Polyhedron from file", fileCreationTime, fileSize)

        # try to load from cache and return with attributes
        loaded = cache.load (fileHash, ".phed")
        if loaded != None:
            return loaded.copy (name = name, attributes = attributes)

        # load stl file
        generated = Polyhedron._loadStlFile (fileName, name, attributes)
        cache.store (generated, fileHash, ".phed")
        return generated
    

    def _calculateFaceNormalVector (self, face):
        """
        Calculates a normal vector for a face.
        """
        p1 = self.points[face[0]]
        p2 = self.points[face[1]]
        p3 = self.points[face[2]]
        v = (p2[0] - p1[0], p2[1] - p1[1], p2[2] - p1[2])
        w = (p3[0] - p1[0], p3[1] - p1[1], p3[2] - p1[2])
        normal = (
            (v[1]*w[2] - v[2]*w[1]),
            (v[2]*w[0] - v[0]*w[2]),
            (v[0]*w[1] - v[1]*w[0])
        )
        return vmath.vsNormalize (normal)


    def writeFileOFF (self, file):
        """
        Write Polyhedron to .off file.
        """
        if isinstance (file, str):
            with open (file, "w") as f:
                self.writeFileOFF (f)
        else:
            file.write ("OFF " + str (len (self.points)) + " " + str (len (self.faces)) + " 0\n")
            for point in self.points:
                file.write (str (point[0]) + " " + str (point[1]) + " " + str (point[2]) + "\n")        
            for face in self.faces:
                file.write ("3 " + str (face[0]) + " " + str (face[1]) + " " + str (face[2]) + "\n")        
    

    def writeFileSTL (self, file):
        """
        Write Polyhedron to .stl file.
        """
        """
        Write Polyhedron to .stl file.
        """
        if isinstance (file, str):
            with open (file, "w") as f:
                self.writeFileSTL (f)
        else:
            # write to stl file
            file.write ("solid Part\n")
            for faceId in range (0, len (self.faces)):
                face = self.faces[faceId]
                n1, n2, n3 = self._calculateFaceNormalVector (face)
                file.write ("  facet normal " + str (n1) + " " + str (n2) + " " + str (n3) + "\n")
                file.write ("    outer loop\n")
                for pointId in range (0, 3):
                    px = self.points[face[pointId]][0]
                    py = self.points[face[pointId]][1]
                    pz = self.points[face[pointId]][2]
                    file.write ("      vertex " + str (px) + " " + str (py) + " " + str (pz) + "\n")
                file.write ("    endloop\n")
                file.write ("  endfacet\n")
            file.write ("endsolid Part\n")


    def _createAdjacentFacePoints (self):
        """
        Create pre computed array of adjacent points.
        """
        adjacent = []
        for faceId in range (0, len (self.points)):
            adjacent.append ([])
        for faceId in range (0, len (self.faces)):
            face = self.faces[faceId]
            for edge in range (0, 3):
                pid = face[edge]
                adjacent[pid].append (faceId)
        return adjacent


    def _createFaceNormals (self):
        """
        Pre calculate face normals.
        """
        normals = []
        for faceIndex in range (0, len (self.faces)):
            normals.append (self._calculateFaceNormalVector (self.faces[faceIndex]))
        return normals


    def _calculateSmoothNormal (self, faceIndex, cornerId, angleThreshold, faceNormals):
        """
        Calculates a smoothed normal for a faces corner point.
        """
        nOwn = faceNormals[faceIndex]
        n = nOwn
        otherFaces = self.getAdjacentFacesOnCorner (faceIndex, cornerId)
        for other in otherFaces:
            nOther = faceNormals[other]
            if vmath.calculateAngle (nOwn, nOther) <= angleThreshold:
                n = vmath.vsAdd (n, nOther)
        return vmath.vsNormalize (n)


    @staticmethod
    def _setFaceNormals (face, n1, n2, n3):
        """
        Create new face by overriding face normals.
        """
        nFace = face[0:3] + [n1, n2, n3]
        if len (face) > 6:
            return nFace + face[6:len (face)]
        return nFace


    def smoothNormals (self, thresholdAngle: float = 30):
        """
        Create a new Polyhedron from this one with auto calculated normal vectors for smooth displaying.
        """

        # try to load from cache
        newHash = hash (self, "smooth", thresholdAngle)
        loaded = cache.load (newHash, ".sphed")
        if loaded != None:
            return loaded.copy (name = self.name, attributes = self.attributes)

        # get face normals
        if not "faceNormals" in self._cache:
            self._cache["faceNormals"] = self._createFaceNormals ()
        faceNormals = self._cache["faceNormals"]

        angleThresholdRadiant = thresholdAngle * math.pi / 180.0
        newNormals = polytool._createIndexedList (self.normals)
        if newNormals == None:
            newNormals = []
        newFaces = []

        # create smoothened normals
        for faceIndex in range (0, len(self.faces)):
            face = self.faces[faceIndex]
            ni1 = None if len (face) < 6 else face[3]
            ni2 = None if len (face) < 6 else face[4]
            ni3 = None if len (face) < 6 else face[5]
            n = self._calculateSmoothNormal (faceIndex, 0, angleThresholdRadiant, faceNormals)
            ni1 = len (newNormals)
            newNormals.append ((n, ni1))
            n = self._calculateSmoothNormal (faceIndex, 1, angleThresholdRadiant, faceNormals)
            ni2 = len (newNormals)
            newNormals.append ((n, ni2))
            n = self._calculateSmoothNormal (faceIndex, 2, angleThresholdRadiant, faceNormals)
            ni3 = len (newNormals)
            newNormals.append ((n, ni3))
            newFaces.append (Polyhedron._setFaceNormals (face, ni1, ni2, ni3))

        # optimize normal vectors
        optNormals, mapNormals = polytool._removeDuplicatesVector3 (newNormals)
        optFaces = []
        for faceIndex in range (0, len(newFaces)):
            face = newFaces[faceIndex]
            n1 = None if len (face) < 6 else face[3]
            n2 = None if len (face) < 6 else face[4]
            n3 = None if len (face) < 6 else face[5]
            nr1 = None if n1 == None else mapNormals[n1]
            nr2 = None if n2 == None else mapNormals[n2]
            nr3 = None if n3 == None else mapNormals[n3]
            optFaces.append (Polyhedron._setFaceNormals (face, nr1, nr2, nr3))

        # create polyhedron from parsed data and store to cache
        generated = Polyhedron (self.points, optFaces, optNormals, self.uvvectors, self.name, self.attributes)
        cache.store (generated, newHash, ".sphed")
        return generated


    def isClosed (self):
        """
        Returns True when the Polygon has a closed surface, False otherwise.
        """
        # value already computed?
        if 'isClosedFlag' in self._cache:
            return self._cache['isClosedFlag']

        # count usage of each edge
        edgeUsageCount = {}
        for face in self.faces:
            for edge in range (0, 3):
                pid1 = face[edge % 3]
                pid2 = face[(edge + 1) % 3]
                pid = str (pid1) + "," + str (pid2) if pid1 < pid2 else str (pid2) + "," + str (pid1)
                if pid in edgeUsageCount:
                    edgeUsageCount[pid] = edgeUsageCount[pid] + 1
                else:
                    edgeUsageCount[pid] = 1

        # compute result
        result = True
        for edge in edgeUsageCount:
            if edgeUsageCount[edge] != 2:
                result = False
                break

        # cache result and return
        self._cache['isClosedFlag'] = result
        return result


    @staticmethod
    def fromUnoptimizedMesh (points, faces, normals = None, uvvectors = None, name: str = None, attributes: Attributes = None):
        """
        Returns a Polygon created from an unoptimized mesh.
        Duplicated points, normals and uvvectors will be removed and the face indices will be adjusted.
        """
        # create optimized version of Polyhedron
        newPoints, newFaces, newNormals, newUvs = Polyhedron._removeDuplicatesWithIndices (
            polytool._createIndexedList (points),
            faces,
            polytool._createIndexedList (normals),
            polytool._createIndexedList (uvvectors),
        )
        return Polyhedron (points = newPoints, faces = newFaces, normals = newNormals, uvvectors = newUvs, name = name, attributes = attributes)


    def getFacesContainingPoint (self, pointId: int):
        """
        Returns a tuple containing the indices of the all faces attached to a point.

        This function shall only be called on closed Polyhedrons, otherwise the behaviour will not be defined.
        """
        # get adjacent face cache
        if not "adjacentFacePoints" in self._cache:
            adjacentFacePoints = self._createAdjacentFacePoints ()
            self._cache["adjacentFacePoints"] = adjacentFacePoints
        else:
            adjacentFacePoints = self._cache["adjacentFacePoints"]

        # return indices of faces connected to pointId
        return adjacentFacePoints[pointId]
        

    def getAdjacentFacesOnCorner (self, faceId: int, cornerId: int):
        """
        Returns a list containg the indices of all adjacent faces connected to a corner of a face.
        This function will only return true adjacent faces, so faceId will not be included in the result.

        This function shall only be called on closed Polyhedrons, otherwise the behaviour will not be defined.
        """
        # get adjacent face cache
        if not "adjacentFacePoints" in self._cache:
            adjacentFacePoints = self._createAdjacentFacePoints ()
            self._cache["adjacentFacePoints"] = adjacentFacePoints
        else:
            adjacentFacePoints = self._cache["adjacentFacePoints"]

        # find adjacent faces on corner
        result = []
        pointId = self.faces[faceId][cornerId % 3]
        for face in adjacentFacePoints[pointId]:
            if face != faceId:
                result.append (face)
        return result


    def getAdjacentFaceOnEdge (self, faceId: int, edgeId: int):
        """
        Returns the index ot the adjacent face on a faces edge.
        Each edge is defined by an edgeId: 0 is the edge between point 0 and point 1, 1 is the edge between point 1 and point 2,
        2 is the edge beween point 2 and point 0. Per definition, a closed non-self-intersecting Polyhederon has exactly one adjacent
        face an edge of a face.

        This function shall only be called on closed Polyhedrons, otherwise the behaviour will not be defined.
        """
        # get adjacent face cache
        if not "adjacentFacePoints" in self._cache:
            adjacentFacePoints = self._createAdjacentFacePoints ()
            self._cache["adjacentFacePoints"] = adjacentFacePoints
        else:
            adjacentFacePoints = self._cache["adjacentFacePoints"]

        # find adjacent face
        pointId1 = self.faces[faceId][edgeId % 3]
        pointId2 = self.faces[faceId][(edgeId + 1) % 3]
        for adjacentFaceIndex in adjacentFacePoints[pointId1]:
            if adjacentFaceIndex != faceId:
                adjacentFace = self.faces[adjacentFaceIndex]
                if (adjacentFace[0] == pointId2) or (adjacentFace[1] == pointId2) or (adjacentFace[2] == pointId2):
                    return adjacentFaceIndex
        return None




class LinearExtrude(tree.Node, Solid):
    """
    Converts a :ref:`Shape<Shape>` into a :ref:`Solid<Solid>` by extruding it along the z axis.

    The x/y coordinates are taken from the child geometry, the resulting :ref:`Solid<Solid>` is centered on the z-axis,
    i.e. z-coordinates are in range [-height / 2 .. height / 2].
    """
    def __init__ (self,
                    children: object = None,
                    height: float = 1,
                    twist: float = 0,
                    scale: float = 1.0,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert child is valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one child"
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
        if not isinstance (children[0], tree.Empty):
            dimensions = 3
            assert children[0].dimensions == 2, "Expected a two dimensional child"
        else:
            dimensions = None
    
        super ().__init__ (children = children, name = name, attributes = attributes)
      
        #: Tuple containing a single child. The child must be a two dimensional csg tree :ref:`Item<Item>`.
        self.children = self.children

        # assert height is valid
        assert isinstance (height, (int, float)), "Parameter height must be a float number"
        assert height > 0, "Parameter height must be greater than zero."
        assert isinstance (twist, (int, float)), "Parameter twist must be a float number"
        assert isinstance (scale, (int, float)), "Parameter scale must be a float number"

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = dimensions

        #: Height of the extruded element.
        self.height = height

        #: Twist in degrees applied along the extrusion.
        self.twist = twist

        #: Scale factor applied along the extrusion. Allows extrusions to be cone shaped.
        self.scale = scale

        # TODO: move this parameter to attributes
        # Number of slices when converting to rasterized shape
        self.slices = None


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'twist',
            'height',
            'scale'
        )




class RotateExtrude(tree.Node, Solid):
    """
    Extrudes a :ref:`Shape<Shape>` to a :ref:`Solid<Solid>` by rotating it around the z-axis.
    """
    def __init__ (self,
                    children: object = None,
                    angle: float = 360,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert child is valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one child"
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
        if not isinstance (children[0], tree.Empty):
            dimensions = 3
            assert children[0].dimensions == 2, "Expected a two dimensional child"
        else:
            dimensions = None

        super ().__init__ (children = children, name = name, attributes = attributes)

        # Number of dimensions is always 3 for :ref:`Solids<Solid>`.
        self.dimensions = dimensions

        #: Tuple containing a single child. The child must be a two dimensional csg tree :ref:`Item<Item>`.
        self.children = self.children

        #: Angle for extrusion in degrees.
        self.angle = angle


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'angle'
        )




# import algorithms
from . import algorithms




def _toPolyhedronsWrapper (item: tree.Item, rasterizingAttributes: tree.Attributes):
    """
    Do final conversion to Ployhedrons.
    """
    # can load from cache?
    cacheHash = hash (item, "as Polyhedron", rasterizingAttributes)
    loaded = cache.load (cacheHash, ".cphed")
    if loaded != None:
        return loaded

    # convert and store to cache
    poly = algorithms.toPolyhedrons (item, rasterizingAttributes)
    cache.store (poly, cacheHash, ".cphed")
    return poly



import os
import subprocess
from shutil import copyfile
from shutil import rmtree
import pathlib
from .external import openscad
from . import hash
from . import cache




class Encoder:
    """
    Base class of video encoder taking single frame images.
    """

    def __init__ (self):
        pass


    def name (self):
        """
        Returns the name of the materializer.
        """
        assert False, "To be implemented by child"


    def assertUsable (self):
        """
        Returns None if the video encoder is usable, an error string if not.
        """
        assert False, "To be implemented by child"


    def presets (self):
        """
        Retuns a list of supported presets.
        """
        assert False, "To be implemented by child"


    def defaultPreset (self):
        """
        Returns the default preset of the video encoder.
        """
        assert False, "To be implemented by child"


    def extension (self):
        """
        Returns the extension of the files generated by the video encoder.
        """
        assert False, "To be implemented by child"


    def encode (self, frames, fps, destinationPath = None, preset = None):
        """
        Generate a video from a sequence of frames.
        """
        assert False, "To be implemented by child"




class _AviH264Encoder (Encoder):
    """
    Video encoder using "mencoder" to produce h264 encoded avi files.
    """
    def __init__ (self):
        super ().__init__ ()
        self.usableChecked = False
        self.usableError = None

    presetMap = {
            'ultrafast': 'ultrafast',
            'superfast': 'superfast',
            'veryfast': 'veryfast',
            'faster': 'faster',
            'fast': 'fast',
            'medium': 'medium',
            'slow': 'slow',
            'slower': 'slower',
            'veryslow': 'veryslow',
            'insane': 'placebo'
    }


    @staticmethod
    def _getMEncoderRunnable ():
        return "mencoder"


    def name (self):
        return "avi-h264"


    def assertUsable (self):
        if self.usableChecked:
            return self.usableError
        try:
            # check if mencoder is runnable
            result = subprocess.run ([_AviH264Encoder._getMEncoderRunnable ()], capture_output = True)
            if result.returncode != 0:
                if not 'MEncoder' in (str(result.stdout) + ' ' + str(result.stderr)):
                    self.usableError = "Can not execute mencoder. Is inside PATH variable?"
            else:
                self.usableError = None
            self.usableChecked = True
            return self.usableError
        except:
            self.usableError = "Can not execute " + _AviH264Encoder._getMEncoderRunnable () + ". Is inside PATH variable?"
        self.usableChecked = True
        return self.usableError
        

    def presets (self):
        return _AviH264Encoder.presetMap.keys ()


    def defaultPreset (self):
        return 'insane'


    def extension (self):
        return 'avi'


    def encode (self, frames, fps, destinationPath = None, preset = None):
        # generate a file name hash by source frames names fps and preset.
        hc = hash ("video", "h264", "avi", fps, preset)
        for frame in frames:
            hc = hash (hc, str (pathlib.Path (frame).resolve ()))

        # check if file is already in cache
        cachePath = pathlib.Path (cache.persistentPath (hc, self.extension ())).resolve ()
        if cachePath.exists ():
            # return cached file or create copy
            if destinationPath == None:
                return str (cachePath)
            else:
                try:
                    copyfile (cachePath, pathlib.Path (destinationPath))
                except:
                    return None
                return str (destinationPath)

        # video doesn't exist, create it...

        # create temp working directory
        tempDir = cache.temporary ()
        os.makedirs (tempDir)

        # get extension of source frames
        framesExtension = None
        for frame in frames:
            fExt = str (frame).split ('.')[-1]
            if framesExtension == None:
                framesExtension = fExt
            else:
                assert fExt == framesExtension, "expected all input frames to have the same extension"

        # symlink video source files into temp directory
        imgId = 0
        for frame in frames:
            # create image index
            imgIdStr = str (imgId)
            while len (imgIdStr) < 6:
                imgIdStr = "0" + imgIdStr
            os.symlink (frame, tempDir + os.path.sep + "img_" + imgIdStr + "." + framesExtension)
            imgId += 1

        # process video rendering in temp directory with sym linked images
        lastDir = os.path.abspath (os.curdir)
        os.chdir (tempDir)

        # temporary avi output
        cacheFileTempAvi = "temp.avi"

        # TODO:
        silent = True

        # prepare command
        command = [_AviH264Encoder._getMEncoderRunnable ()]
        command.append ("mf://img_*.png")
        command.append ("-mf")
        command.append ("fps=" + str (fps) + ":type=png")
        command.append ("-oac")
        command.append ("copy")
        command.append ("-ovc")
        command.append ("x264")
        command.append ("-x264encopts")
        command.append ("preset=" + _AviH264Encoder.presetMap[preset] + ":tune=animation:threads=auto")
        command.append ("-o")
        command.append (cacheFileTempAvi)

        # do compression
        result = subprocess.run (command, capture_output=silent)
        if result.returncode != 0:
            if silent:
                print (result.stderr)
                print (result.stdout)
            try:
                os.chdir (lastDir)
                rmtree (tempDir)
            except:
                pass
            return None

        # copy to cache
        cacheFile = cache.persistentPath (hc, self.extension (), True)
        os.rename (cacheFileTempAvi, cacheFile)

        # leave and remove temporary directory
        os.chdir (lastDir)
        try:
            rmtree (tempDir)
        except:
            pass

        # need to copy to output file?
        if destinationPath == None:
            return str (cacheFile)
        else:
            try:
                copyfile (cacheFile, pathlib.Path (destinationPath))
            except:
                return None
            return str (destinationPath)

"""
Static instance of AVI-H264 Encoder.
"""
AviH264 = _AviH264Encoder ()




class _MP4H264Encoder (Encoder):
    """
    MP4 encoder using AviH264 encoder.
    """

    def __init__ (self):
        super ().__init__ ()
        self.usableChecked = False
        self.usableError = None


    def name (self):
        return "mp4-h264"


    @staticmethod
    def _getMP4BoxRunnable ():
        return "MP4Box"


    def assertUsable (self):
        """
        Returns None if the video encoder is usable, an error string if not.
        """
        if self.usableChecked:
            return self.usableError
        try:
            # check if parent encoder is runnable
            aviErr = AviH264.assertUsable ()
            if aviErr != None:
                self.usableChecked = True
                self.usableError = aviErr
                return self.usableError

            # check if mencoder is runnable
            result = subprocess.run ([_MP4H264Encoder._getMP4BoxRunnable ()], capture_output = True)
            if result.returncode != 0:
                if not 'MP4Box' in (str(result.stdout) + ' ' + str(result.stderr)):
                    self.usableError = "Can not execute MP4Box. Is inside PATH variable?"
            else:
                self.usableError = None
            self.usableChecked = True
            return self.usableError
        except:
            self.usableError = "Can not execute " + _MP4H264Encoder._getMP4BoxRunnable () + ". Is inside PATH variable?"
        self.usableChecked = True
        return self.usableError


    def presets (self):
        """
        Retuns a list of supported presets.
        """
        return AviH264.presets ()


    def defaultPreset (self):
        """
        Returns the default preset of the video encoder.
        """
        return AviH264.defaultPreset ()


    def extension (self):
        """
        Returns the extension of the files generated by the video encoder.
        """
        return 'mp4'


    def encode (self, frames, fps, destinationPath = None, preset = None):
        """
        Generate a video from a sequence of frames.
        """
        # generate a file name hash by source frames names fps and preset.
        hc = hash ("video", "h264", "mp4", fps, preset)
        for frame in frames:
            hc = hash (hc, str (pathlib.Path (frame).resolve ()))

        # check if file is already in cache
        cachePath = pathlib.Path (cache.persistentPath (hc, self.extension ())).resolve ()
        if cachePath.exists ():
            # return cached file or create copy
            if destinationPath == None:
                return str (cachePath)
            else:
                try:
                    copyfile (cachePath, pathlib.Path (destinationPath))
                except:
                    return None
                return str (destinationPath)

        # video doesn't exist, create it...

        # Encode via parent encoder (get avi file path)
        preEncoded = AviH264.encode (frames, fps, None, preset)

        # create temp working directory
        tempDir = cache.temporary ()
        os.makedirs (tempDir)

        # symlink video into temporary directory
        os.symlink (preEncoded, tempDir + os.path.sep + 'input.avi')

        # process inside temporary directory
        lastDir = os.path.abspath (os.curdir)
        os.chdir (tempDir)

        # TODO:
        silent = True

        # unpack h264 stream
        unpackCommand = [_MP4H264Encoder._getMP4BoxRunnable (), "-aviraw", "video", 'input.avi']
        result = subprocess.run (unpackCommand, capture_output=silent)
        if result.returncode != 0:
            if silent:
                print (result.stderr)
                print (result.stdout)
            try:
                os.chdir (lastDir)
                rmtree (tempDir)
            except:
                pass
            return None

        # temporary output file
        cacheFileTemp = "output.mp4"

        # pack mp4 file
        packCommand = [_MP4H264Encoder._getMP4BoxRunnable (), "-add", "input_video.h264", cacheFileTemp]
        result = subprocess.run (packCommand, capture_output=silent)
        if result.returncode != 0:
            if silent:
                print (result.stderr)
                print (result.stdout)
            try:
                os.chdir (lastDir)
                rmtree (tempDir)
            except:
                pass
            return None

        # copy to cache
        cacheFile = cache.persistentPath (hc, self.extension (), True)
        os.rename (cacheFileTemp, cacheFile)

        # leave & remove temporary directory
        try:
            os.chdir (lastDir)
            rmtree (tempDir)
        except:
            pass

        # need to copy to output file?
        if destinationPath == None:
            return str (cacheFile)
        else:
            try:
                copyfile (cacheFile, pathlib.Path (destinationPath))
            except:
                return None
            return str (destinationPath)

"""
Static instance of MP4-H264 Encoder.
"""
MP4H264 = _MP4H264Encoder ()




"""
List of registered encoders.
"""
_registeredEncoders = [MP4H264, AviH264]




def getEncoders ():
    """
    Returns a list of all registered video encoder instances.
    """
    return _registeredEncoders




def getEncoder (name):
    """
    Returns an encoder by name or None if not found.
    """
    for m in _registeredEncoders:
        if m.name () == name:
            return m
    return None




def registerEncoder (encoder):
    """
    Registers an video encoder at the registry.
    """
    assert False, "TODO:"




def parseSettings (formatStr):
    """
    Parses a format and returns a tuple (encoderInstance, presetName). When parsing failed, encoderInstance will be None.
    """
    # split string
    fmt = formatStr.split ('.')
    if len (fmt) < 0:
        return (None, None)
    
    # find encoder
    encoderInstance = None
    for e in _registeredEncoders:
        if len (fmt) > 0:
            if e.name () == fmt[0]:
                encoderInstance = e
                break
        else:
            encoderInstance = _registeredEncoders[0] if len (_registeredEncoders) > 0 else None

    # return tuple with parsed settings
    presetName = '.'.join (fmt[1:])
    if presetName in (None, ""):
        if encoderInstance != None:
            presetName = encoderInstance.defaultPreset ()
    return (encoderInstance, presetName)




def extension (formatStr):
    """
    Get file extension for rendering format.
    """
    assert False, "TODO:"



def Material (colorVector):
    """
    Material color.
    """
    if colorVector == None:
        return None
    else:
        assert isinstance (colorVector, (list, tuple))
        assert len (colorVector), 3
        for cid in range (0, 3):
            assert colorVector[cid] >= 0
            assert colorVector[cid] <= 1
        return colorVector



from . import hash as hashUtil


def hash(*arguments):
    """
    Calculates a hash code for a list of arguments with arbirtary types, returns an unsigned 64 bit number.
    """
    value = ""
    for a in arguments:
        value = value + str (hashUtil._hashArgument (a)) + ","
    return hashUtil._hashString (value)




def splitLines (text: str) -> list:
    """
    Split string into single lines, retaining also empty ones.
    """
    lastEntry = 0
    elements = []
    strLen = len (text)
    for i in range (0, strLen):
        if text[i] == "\n":
            elements.append (text[lastEntry : i])
            lastEntry = i + 1
    if len (text) >= lastEntry:
        elements.append (text[lastEntry : len (text)])
    return elements




def breakLines (text, maxWidth, maxWidthFirst = None):
    """
    Breaks a line into multiple lines.
    """
    if maxWidthFirst == None:
        maxWidthFirst = maxWidth
    mw = maxWidthFirst
    result = []
    for l in splitLines (text):
        temp = l
        while len (temp) > mw:
            pos = _breakLineSplitPos (temp, mw)
            result.append (temp[0 : pos])
            if pos < len (temp):
                if temp[pos] == " " or temp[pos] == "\n" or temp[pos] == "\t":
                    pos = pos + 1
                temp = temp [pos : len (temp)]
            else:
                temp = ""
            mw = maxWidth
        result.append (temp)
    return result





def _breakLineSplitPos (text, maxWidth):
    """
    Get position for text splitting.
    """
    if len (text) < maxWidth:
        return len (text)
    lastSplitPos = None
    for i in range (1, len (text)):
        if text[i] == " " or text[i] == "\t" or text[i] == "\n":
            if (i >= maxWidth):
                if lastSplitPos == None:
                    return maxWidth
                else:
                    return lastSplitPos
            lastSplitPos = i
    if lastSplitPos == None:
        return maxWidth
    else:
        return lastSplitPos





def formatTableEntryLeftSize (left, width, leftSize, centerSpace = 2, indent = 2):
    """
    Get size of left column after formatting table entry.
    """
    lbs = leftSize + centerSpace
    if (lbs + 2) > width:
        width = (lbs + 2)
    leftSplit = breakLines (left, leftSize - indent, leftSize)
    maxLineSize = 0
    for i in range (0, len (leftSplit)):
        lt = leftSplit[i]
        if i > 0:
            lt = (" " * indent) + lt
        if len (lt) > maxLineSize:
            maxLineSize = len (lt)
    return maxLineSize





def formatTextBlock (text, indent = 2, maxLineSize = None, prefix: str = ""):
    """
    Format block of Text.
    """
    mls = 1000
    if maxLineSize != None:
        mls = maxLineSize
    breakSizeFirst = mls - len (prefix)
    breakSize = breakSizeFirst - indent
    splited = breakLines (text, breakSize, breakSizeFirst)
    indentStr = " " * indent
    buffer = ""
    for line in splited:
        if buffer == "":
            buffer = prefix + line
        else:
            buffer = buffer + "\n" + indentStr + prefix + line
    return buffer




def formatTableEntry (left, right, width = 80, leftSize = 30, centerSpace = 2, indent = 2, prefix = ""):
    """
    Format table entry.
    """
    lbs = len (prefix) + leftSize + centerSpace
    if (lbs + 2) > width:
        width = (lbs + 2)
    rbs = width - lbs
    leftSplit = breakLines (left, leftSize - indent, leftSize)
    rightSplit = breakLines (right, rbs - indent, rbs)
    while len (leftSplit) < len (rightSplit):
        leftSplit.append ("")
    while len (leftSplit) > len (rightSplit):
        rightSplit.append ("")
    buffer = ""
    for i in range (0, len (leftSplit)):
        lt = leftSplit[i]
        rt = rightSplit[i]
        if i > 0:
            lt = (" " * indent) + lt
            rt = (" " * indent) + rt
        pads = lbs - len (lt)
        if (pads > 0):
            out = lt + " " * pads + rt
        else:
            out = lt + rt
        if buffer != "":
            buffer = buffer + "\n"
        buffer = buffer + prefix + out
    return buffer


import os
import pathlib
import subprocess
from ... import tree
from ... import shape
from ... import solid
from ... import transform
from ... import boolean
from ... import algorithms
from ... import attributeTypes
from .. import textwriter
from .. import hash
from .. import cache
from .. import downsample




class _TreeIgnoreFilter (algorithms.RemoveNodes):
    """
    Tree ignore filter for removing nodes ignored by OpenScad.
    """
    def __init__ (self):
        super ().__init__ ()

        # node removal filter
        self.rule (tree.Empty, self.enterFilter, self.leaveFilter)

        # TODO: Add attribute '_openscad.material' to each item.
        # This information will be used by the OpenScad tree exporter to determine a rendering color.

"""
Single instance of tree ignore filter.
"""
TreeIgnoreFilter = _TreeIgnoreFilter ()




class _TreeReplaceNonTrivialTypes (tree.Visitor):
    """
    Replaces non trivial tree nodes before rendering.
    """
    def __init__ (self):
        super ().__init__ ()

        # replacement rules
        self.rule (shape.Bezier, leave = self.leaveCurve)
        self.rule (shape.Polar, leave = self.leaveCurve)
        self.rule (shape.Complex, leave = self.leaveComplex)


    def execute (self, item, attributes):
        """
        Collect render attributes for node.
        """
        context = tree.Visitor.Context (attributes)
        return self.visit (item, context)


    def leaveCurve (self, item, context):
        """
        Convert curve to rasterized Polygon
        """
        rasterizingAttributes = context.resolvedAttributes[-1]
        return shape.Polygon.fromShape (item, rasterizingAttributes = rasterizingAttributes, name = item.name, attributes = item.attributes)


    def leaveComplex (self, item, context):
        """
        Convert complex to rasterized Polygon
        """
        rasterizingAttributes = context.resolvedAttributes[-1]
        return item.rasterize (rasterizingAttributes, False)

"""
Single instance of replace non trivial types filter.
"""
TreeReplaceNonTrivialTypes = _TreeReplaceNonTrivialTypes ()




def TreeCollectRenderAttributes (item):
    """
    Collect render attributes for tree node.
    """ 
    # already tagged with a material?   
    if item == None:
        return None
    matAttr = item.attributes.get ('_openscad.material')
    if matAttr != None:
        return item

    matAttr = item.attributes.get ('material')
    if matAttr != None:
        # has own material definition
        if isinstance (item, tree.Node):
            newChildren = []
            for child in item.children:
                newChildren.append (TreeCollectRenderAttributes (child))
            return item.copy (children = newChildren, attributes = item.attributes.override ({'_openscad.material': matAttr}))
        else:
            return item.copy (attributes = item.attributes.override ({'_openscad.material': matAttr}))
    else:
        # get material definition from children
        nAttributes = item.attributes

        # recurse on children
        if isinstance (item, tree.Node):
            newChildren = []
            for child in item.children:
                newChildren.append (TreeCollectRenderAttributes (child))

        if isinstance (item, (solid.LinearExtrude, solid.RotateExtrude, transform.Transform)):
            childAttr = newChildren[0].attributes.get ('_openscad.material')
            if childAttr != None:
                nAttributes = nAttributes.override ({'_openscad.material': childAttr})
        elif isinstance (item, boolean.Union):
            childAttr = None
            mismatch = False
            for child in newChildren:
                tChildAttr = child.attributes.get ('_openscad.material')
                if childAttr != None:
                    if childAttr != tChildAttr:
                        mismatch = True
                    else:
                        childAttr = tChildAttr
            if (mismatch == False) and (childAttr != None):
                nAttributes = nAttributes.override ({'_openscad.material': childAttr})
        elif isinstance (item, (boolean.Difference, boolean.Intersection)):
            childAttr1 = newChildren[0].attributes.get ('_openscad.material')
            childAttr2 = newChildren[1].attributes.get ('_openscad.material')
            if childAttr1 == None:
                if childAttr2 != None:
                    nAttributes = nAttributes.override ({'_openscad.material': childAttr2})
            else:
                nAttributes = nAttributes.override ({'_openscad.material': childAttr1})

        # rebuild node
        if isinstance (item, tree.Node):
            return item.copy (children = newChildren, attributes = nAttributes)
        else:
            return item.copy (attributes = nAttributes)




class _TreeVisitor (tree.Visitor):
    """
    A tree visitor writing an OpenScad script.
    """
    def __init__ (self):
        super ().__init__ ()

        # handle two dimensional basic shapes
        self.rule (shape.Circle, self.enterCircle, self.leaveGeometry)
        self.rule (shape.Square, self.enterSquare, self.leaveGeometry)
        self.rule (shape.Polygon, self.enterPolygon, self.leaveGeometry)
        self.rule (shape.Projection, self.enterProjection, self.leaveProjection)

        # handle three dimensional basic shapes
        self.rule (solid.Sphere, self.enterSphere, self.leaveGeometry)
        self.rule (solid.Cube, self.enterCube, self.leaveGeometry)
        self.rule (solid.Cylinder, self.enterCylinder, self.leaveGeometry)
        self.rule (solid.Polyhedron, self.enterPolyhedron, self.leaveGeometry)
        self.rule (solid.LinearExtrude, self.enterLinearExtrude, self.leaveExtrude)
        self.rule (solid.RotateExtrude, self.enterRotateExtrude, self.leaveExtrude)

        # handle transformations
        self.rule (transform.Translate, self.enterTranslate, self.leaveTransform)
        self.rule (transform.Scale, self.enterScale, self.leaveTransform)
        self.rule (transform.Rotate, self.enterRotate, self.leaveTransform)

        # handle boolean operations
        self.rule (boolean.Union, self.enterUnion, self.leaveUnion)
        self.rule (boolean.Difference, self.enterDifference, self.leaveBoolean)
        self.rule (boolean.Intersection, self.enterIntersection, self.leaveBoolean)

        # handle empty node
        self.rule (tree.Empty, self.enterIgnore, self.leaveIgnore)

        # handle part node
        self.rule (tree.Part, self.enterPart, self.leavePart)


    class Context (tree.Visitor.Context):
        """
        Own context class.
        """
        def __init__ (self, attributes, writer):
            super ().__init__ (attributes)
            self.writer = writer


    def execute (self, item: tree.Item, attributes: dict, writer: textwriter.TextWriter):
        """
        Write a tree as a OpenScad script.
        """
        # Remove unrenderable items, replace non trivial item types, translate groups into unions and remove empty nodes.
        filtered = TreeIgnoreFilter.execute (item, attributes)                      # remove ignored rendering items
        filtered = TreeReplaceNonTrivialTypes.execute (filtered, attributes)        # replace none trivial render items
        filtered = algorithms.groupsToUnions (filtered)                             # rewite groups as unions.
        filtered = algorithms.removeEmpty (filtered)                                # purge empty nodes from item.
        filtered = algorithms.pushTransformsFurtherInside (filtered)                # push transforms to be most inner
        filtered = algorithms.pushExtrusionsFurtherInside (filtered)                # push extrusions to be most inner
        filtered = TreeCollectRenderAttributes (filtered)                           # collect render colors

        # flatten to OpenScad script
        context = _TreeVisitor.Context (attributes, writer)
        self.visit (filtered, context)


    def enterPart (self, item, context):
        """
        Enter tree.Part element.
        """
        return True


    def leavePart (self, item, context):
        """
        Leave tree.Part element.
        """
        return item


    def enterIgnore (self, item, context):
        """
        Ignore item on tree traversal.
        """
        return False


    def leaveIgnore (self, item, context):
        """
        Ignore item on tree traversal.
        """
        return item


    def defaultEnter (self, item, context):
        """
        Handler function for entering non handled types.
        """
        assert False, "Item class not handled."


    def defaultLeave (self, item, context):
        """
        Handler function for leaving non handled types.
        """
        assert False, "Item class not handled."


    def writeComment (self, item, context):
        """
        Write comment for node.
        """
        if item.name != None:
            if item.name != "":
                context.writer.write ("// " + item.name + ":\n")


    def writeAttributes (self, item, context):
        """
        Write attributes for node.
        """
        attributes = context.resolvedAttributes[-1].attributes
        material = attributes['_openscad.material'] if '_openscad.material' in attributes else None
        if material != None:
            context.writer.write ("color ([")
            context.writer.write (str (material[0]))
            context.writer.write (", " + str (material[1]))
            context.writer.write (", " + str (material[2]))
            if len (material) > 3:
                context.writer.write (", " + str (material[3]))
            context.writer.write ("]) \n")


    def endAttributes (self, item, context):
        """
        End processing item with attributes.
        """
        pass


    def formatTesselationParameters (self, item, context):
        """
        Calculates the tesselation parameter string.
        """
        # get rasterizing settings
        minAngle, minSize, fixedCount, minSlices = context.resolvedAttributes[-1].rasterizingAttributes ()
        result = ""
        if fixedCount != None:
            result = ", $fn = " + context.writer.formatNumber (fixedCount)
        else:
            if minAngle != None:
                result += ", $fa = " + context.writer.formatNumber (minAngle)
            if minSize != None:
                result += ", $fs = " + context.writer.formatNumber (minSize)
        return result


    def enterCircle (self, item, context):
        """
        Write a shape.Circle instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        tes = self.formatTesselationParameters (item, context)
        context.writer.write ("circle (r = ")
        context.writer.writeNumber (item.radius)
        context.writer.write (tes + ");\n")
        return True


    def enterSquare (self, item, context):
        """
        Write a shape.Square instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("square (size = [")
        context.writer.writeNumber (item.size[0])
        context.writer.write (", ")
        context.writer.writeNumber (item.size[1])
        context.writer.write ("], center = true);\n")
        return True
        

    def writePolygonPath (self, item, path, context):
        """
        Format a polygon path.
        """
        context.writer.write ("[")
        for pid in range (0, len (path)):
            if pid == (len (path) - 1):
                context.writer.write (str (path[pid]))
            else:
                context.writer.write (str (path[pid]) + ", ")
        context.writer.write ("]")


    def enterPolygon (self, item, context):
        """
        Enter a ploygon.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("polygon (points = [")
        context.writer.increaseIndent ()
        self.writePoints2D (item.points, context)
        context.writer.write ("],\n")
        context.writer.decreaseIndent ()
        context.writer.write ("paths = [\n")
        context.writer.increaseIndent ()
        self.writePolygonPath (item, item.outer, context)
        for cutout in item.cutouts:
            context.writer.write (", \n")
            self.writePolygonPath (item, cutout, context)
        context.writer.write ("]);\n")
        context.writer.decreaseIndent ()
        return True
        

    def enterProjection (self, item, context):
        """
        Enter projection.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("projection (cut = ")
        context.writer.write ("true" if item.cut == True else "false")
        context.writer.write (")\n")
        context.writer.increaseIndent ()
        return True


    def leaveProjection (self, item, context):
        """
        Leave projection.
        """
        self.endAttributes (item, context)
        context.writer.decreaseIndent ()
        return item
        

    def enterSphere (self, item, context):
        """
        Write a solid.Sphere instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        tes = self.formatTesselationParameters (item, context)
        context.writer.write ("sphere (r = ")
        context.writer.writeNumber (item.radius)
        context.writer.write (", center = true" + tes + ");\n")
        return True
        

    def enterCube (self, item, context):
        """
        Write a solid.Cube instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("cube (size = [")
        context.writer.writeNumber (item.size[0])
        context.writer.write (", ")
        context.writer.writeNumber (item.size[1])
        context.writer.write (", ")
        context.writer.writeNumber (item.size[2])
        context.writer.write ("], center = true);\n")
        return True
        

    def enterCylinder (self, item, context):
        """
        Write a solid.Cylinder instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        tes = self.formatTesselationParameters (item, context)
        context.writer.write ("cylinder (h = ")
        context.writer.writeNumber (item.height)
        context.writer.write (", r1 = ")
        context.writer.writeNumber (item.radius1)
        context.writer.write (", r2 = ")
        context.writer.writeNumber (item.radius2)
        context.writer.write (", center = true" + tes + ");\n")
        return True
        

    def writePoints2D (self, points, context):
        """
        Write list of points.
        """
        for pointId in range (0, len (points) - 1):
            point = points[pointId]
            context.writer.write ("[")
            context.writer.writeNumber (point[0])
            context.writer.write (", ")
            context.writer.writeNumber (point[1])
            context.writer.write ("],\n")
        if len (points) > 0:
            point = points[-1]
            context.writer.write ("[")
            context.writer.writeNumber (point[0])
            context.writer.write (", ")
            context.writer.writeNumber (point[1])
            context.writer.write ("]\n")
        

    def writePoints3D (self, points, context):
        """
        Write list of points.
        """
        for pointId in range (0, len (points) - 1):
            point = points[pointId]
            context.writer.write ("[")
            context.writer.writeNumber (point[0])
            context.writer.write (", ")
            context.writer.writeNumber (point[1])
            context.writer.write (", ")
            context.writer.writeNumber (point[2])
            context.writer.write ("],\n")
        if len (points) > 0:
            point = points[-1]
            context.writer.write ("[")
            context.writer.writeNumber (point[0])
            context.writer.write (", ")
            context.writer.writeNumber (point[1])
            context.writer.write (", ")
            context.writer.writeNumber (point[2])
            context.writer.write ("]\n")
        

    def writeFaceIds (self, faces, context):
        """
        Write list of faces.
        """
        for faceId in range (0, len (faces) - 1):
            face = faces[faceId]
            context.writer.write ("[")
            context.writer.write (str (face[0]))
            context.writer.write (", ")
            context.writer.write (str (face[1]))
            context.writer.write (", ")
            context.writer.write (str (face[2]))
            context.writer.write ("],\n")
        if len (faces) > 0:
            face = faces[-1]
            context.writer.write ("[")
            context.writer.write (str (face[0]))
            context.writer.write (", ")
            context.writer.write (str (face[1]))
            context.writer.write (", ")
            context.writer.write (str (face[2]))
            context.writer.write ("]\n")


    def enterPolyhedron (self, item, context):
        """
        Write a polyhedron to OpenScad script.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("polyhedron (points = [\n")
        context.writer.increaseIndent ()
        self.writePoints3D (item.points, context)
        context.writer.write ("],\n")
        context.writer.decreaseIndent ()
        context.writer.write ("faces = [\n")
        context.writer.increaseIndent ()
        self.writeFaceIds (item.faces, context)
        context.writer.write ("]);\n")
        context.writer.decreaseIndent ()
        return True
        

    def enterLinearExtrude (self, item, context):
        """
        Enter linear extrusion.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        tes = self.formatTesselationParameters (item, context)
        context.writer.write ("linear_extrude (height = ")        
        context.writer.writeNumber (item.height)
        if item.twist != 0:
            context.writer.write (", twist = ")
            context.writer.writeNumber (item.twist)
        if item.scale != 1:
            context.writer.write (", scale = ")
            context.writer.writeNumber (item.scale)
        context.writer.write (", convexity = 10")
        minAngle, minSize, fixedCount, minSlices = context.resolvedAttributes[-1].rasterizingAttributes ()
        if minSlices != None:
            context.writer.write (", slices = ")
            context.writer.writeNumber (minSlices)            
        context.writer.write (", center = true" + tes + ")\n")
        context.writer.increaseIndent ()
        return True
        

    def enterRotateExtrude (self, item, context):
        """
        Enter linear extrusion.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        tes = self.formatTesselationParameters (item, context)
        context.writer.write ("rotate_extrude (angle = ")        
        context.writer.writeNumber (item.angle)
        context.writer.write (", convexity = 10")
        context.writer.write (tes + ")\n")
        context.writer.increaseIndent ()
        return True


    def leaveExtrude (self, item, context):
        """
        Leave an extrusion.
        """
        self.endAttributes (item, context)
        context.writer.decreaseIndent ()
        return item


    def leaveGeometry (self, item, context):
        """
        Leave a geometry item.
        """
        self.endAttributes (item, context)
        return item


    def enterTranslate (self, item, context):
        """
        Enter translation.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("translate ([")        
        context.writer.writeNumber (item.x)
        context.writer.write (", ")
        context.writer.writeNumber (item.y)
        context.writer.write (", ")
        context.writer.writeNumber (item.z)
        context.writer.write ("])\n")
        context.writer.increaseIndent ()
        return True
        

    def enterScale (self, item, context):
        """
        Enter scale.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("scale ([")        
        context.writer.writeNumber (item.sx)
        context.writer.write (", ")
        context.writer.writeNumber (item.sy)
        context.writer.write (", ")
        context.writer.writeNumber (item.sz)
        context.writer.write ("])\n")
        context.writer.increaseIndent ()
        return True
        

    def enterRotate (self, item, context):
        """
        Enter rotate.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("rotate ([")        
        context.writer.writeNumber (item.rx)
        context.writer.write (", ")
        context.writer.writeNumber (item.ry)
        context.writer.write (", ")
        context.writer.writeNumber (item.rz)
        context.writer.write ("])\n")
        context.writer.increaseIndent ()
        return True


    def leaveTransform (self, item, context):
        """
        Leave transform.
        """
        self.endAttributes (item, context)
        context.writer.decreaseIndent ()
        return item


    def enterUnion (self, item, context):
        """
        Write a boolean.Union instance.
        """
        if len (context.stack) > 1:
            self.writeComment (item, context)
            self.writeAttributes (item, context)
            context.writer.write ("union () {\n")        
            context.writer.increaseIndent ()
        return True
        

    def enterDifference (self, item, context):
        """
        Write a boolean.Difference instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("difference () {\n")        
        context.writer.increaseIndent ()
        return True
        

    def enterIntersection (self, item, context):
        """
        Write a boolean.Intersection instance.
        """
        self.writeComment (item, context)
        self.writeAttributes (item, context)
        context.writer.write ("intersection () {\n")        
        context.writer.increaseIndent ()
        return True
        

    def leaveBoolean (self, item, context):
        """
        Leave boolean operation.
        """
        self.endAttributes (item, context)
        context.writer.decreaseIndent ()
        context.writer.write ("};\n")
        return item
        

    def leaveUnion (self, item, context):
        """
        Leave union operation.
        """
        if len (context.stack) > 1:
            self.endAttributes (item, context)
            context.writer.decreaseIndent ()
            context.writer.write ("};\n")
        return item
            
"""
Single instance of tree visitor.
"""
TreeVisitor = _TreeVisitor ()




# Open scad toolkit class
class _OpenScad:
    pass

_openScadInstance = None


def _writeScript (item: tree.Item, fileName: str, attributes: tree.Attributes = None):
    """
    Write a csg tree to a OpenScad script file. The file will be overwritten.
    """
    writer = textwriter.FileWriter (fileName)
    fmtWriter = textwriter.TextWriter (writer)
    if attributes == None:
        attributes = tree.Attributes.defaults ()
    TreeVisitor.execute (item, attributes, fmtWriter)
    fmtWriter.flush ()
    fmtWriter.close ()
    writer.flush ()
    writer.close ()


def getOpenScadRunnable ():
    """
    Get OpenScad runnable.
    """
    return "openscad"


def getScadScript (item: tree.Item, attributes: tree.Attributes = None):
    """
    Get the absolute path to an OpenScad script describing the geometry of *item* with *attributs*.
    """
    if attributes == None:
        attributes = tree.Attributes.defaults ()
    hc = hash ('openscad', item, attributes)
    persistent = cache.persistentPath (hc, 'scad', True)
    if pathlib.Path (persistent).exists ():
        return persistent
    tempFile = cache.temporary ('scad')
    _writeScript (item, tempFile, attributes)
    try:
        pathlib.os.rename (tempFile, persistent)
    except:
        pass
    return persistent


def getMaterialized (fmt: str, item: tree.Item, attributes: tree.Attributes = None):
    """
    Get materialized OpenScad file by format.
    """
    # try to get from cache
    assert fmt in ('off', 'stl'), "Unknown materializeing format for OpenScad"
    hc = hash ('materialized', item, attributes)
    persistent = cache.persistentPath (hc, fmt, True)
    if pathlib.Path (persistent).exists ():
        return persistent
    tempFile = cache.temporary (fmt)

    # TODO: silent flag
    silent = True

    # render to temporary file
    script = getScadScript (item, attributes)
    cmd = [getOpenScadRunnable (), script, '-o', str (tempFile)]
    result = subprocess.run (cmd, capture_output = silent)
    if result.returncode == 0:
        # move to cache
        try:
            pathlib.os.rename (tempFile, persistent)
        except:
            return None
        return persistent
    else:
        # handle error
        if silent:
            print (result.stderr)
            print (result.stdout)
        return None


def getRendering (fmt: str, item: tree.Item, attributes: tree.Attributes = None):
    """
    Get rendered OpenScad file by format.
    """
    assert fmt in ('png', 'svg'), "Unknown redering format for OpenScad"
    hc = hash ('rendered', item, attributes)
    persistent = cache.persistentPath (hc, fmt, True)
    if pathlib.Path (persistent).exists ():
        return persistent
    tempFile = cache.temporary (fmt)

    renderAttrs = attributes.renderAttributes ()
    if not fmt in ('png'):
        renderAttrs['antialias'] = 1

    # assert rendering options are valid
    assert isinstance (renderAttrs['width'], (int, float))
    assert renderAttrs['width'] > 0
    assert isinstance (renderAttrs['height'], (int, float))
    assert renderAttrs['height'] > 0
    assert isinstance (renderAttrs['antialias'], (int, float))
    assert renderAttrs['antialias'] > 0
    assert isinstance (renderAttrs['view.axis'], bool)
    assert isinstance (renderAttrs['view.scales'], bool)
    assert isinstance (renderAttrs['colorScheme'], str)

    antialias = renderAttrs['antialias']

    # compute rendering options
    width = renderAttrs['width'] * antialias
    height = renderAttrs['height'] * antialias
    renderOptions = ['--imgsize=' + str (width) + "," + str (height)]
    if renderAttrs['view.axis'] or renderAttrs['view.scales'] or renderAttrs['view.crosshairs']:
        vstr = ""
        if renderAttrs['view.axis']:
            vstr += 'axes'
        if renderAttrs['view.scales']:
            if vstr != "":
                vstr += ","
            vstr += 'scales'
        if renderAttrs['view.crosshairs']:
            if vstr != "":
                vstr += ","
            vstr += 'crosshairs'
        renderOptions += ['--view=' + vstr]
    if not renderAttrs['colorScheme'] in ('', None):
        renderOptions += ['--colorscheme=' + renderAttrs['colorScheme']]

    # process camera & perspective settings
    cameraAttribs = attributes.cameraAttributes ()
    renderOptions += ['--camera=' + ','.join (map (str, cameraAttribs['view']))]
    if cameraAttribs['projection'] == 'orthogonal':
        renderOptions += ['--projection=o']

    # TODO: silent flag
    silent = True

    # render to temporary file
    script = getScadScript (item, attributes)
    cmd = [getOpenScadRunnable (), script, '-o', str (tempFile)]
    cmd = cmd + renderOptions
    result = subprocess.run (cmd, capture_output = silent)
    
    # down sample anti aliased image
    if antialias > 1:
        downsample.downsampleImage (tempFile, tempFile, antialias)

    if result.returncode == 0:
        # move to cache
        try:
            pathlib.os.rename (tempFile, persistent)
        except:
            return None
        return persistent
    else:
        # handle error
        if silent:
            print (result.stderr)
            print (result.stdout)
        return None
